function [min_eig, max_eig] = eigk(A, n, opts)
%function [min_eig, max_eig] = eigk(A, n, opts)
%
% EIGK computes the smallest and the largest eigenvelue of A. Here
% A is in an operator form (a R^n to R^n function). EIGK uses the
% Lanczos algorithm for a symmetric matrix.
%
% default values:
%
%   opts.maxiter     600
%   opts.tol        1e-4
%   opts.t_start      20
%   opts.t_step       20
%

if ~exist('opts','var')      opts = struct;        end
if ~isfield(opts,'maxiter')  opts.maxiter = 600;   end
if ~isfield(opts,'tol')      opts.tol     = 1e-4;  end
if ~isfield(opts,'t_start')  opts.t_start = 20;    end
if ~isfield(opts,'t_step')   opts.t_step  = 20;    end

maxiter = opts.maxiter;
tol     = opts.tol;
t_step  = opts.t_step;
t_start = opts.t_start;

beta    = zeros(maxiter+1,1);
alpha   = zeros(maxiter,1);
Q       = zeros(n,maxiter+1);
qi_1    = zeros(n,1);
T       = zeros(maxiter+1);

Q(:,1)  = randn(n,1);
Q(:,1)  = Q(:,1)/norm(Q(:,1));

i = 1;
min_eig_prev = -1;
max_eig_prev = -1;
while i <= maxiter
  %--lanczos step
  qi = Q(:,i);
  w = A(qi) - beta(i)*qi_1;
  alpha(i) = w'*qi;
  w = w - alpha(i)*qi;
  %%----------------------
  %%  re-orthogonalization
  w = w - Q(:,1:i-1)*(w'*Q(:,1:i-1))';
  %%----------------------
  beta(i+1) = norm(w);
  Q(:,i+1) = w/beta(i+1);
  qi_1 = qi;
  iter = i;
  
  T(i,i) = alpha(i);
  T(i,i+1) = beta(i+1);
  T(i+1,i) = beta(i+1);
  
  %--convergence test
  if (i>=t_start && mod(i,t_step)==0) || (i==maxiter)
    eig_val = eig(T(1:iter,1:iter));
    min_eig = min(eig_val);
    max_eig = max(eig_val);
    if abs(min_eig-min_eig_prev) < tol & ...
          abs(max_eig-max_eig_prev) < tol
      break;
    else
      min_eig_prev = min_eig;
      max_eig_prev = max_eig;
    end
  end
  
  i = i + 1;
end

if i > maxiter && i ~= n+1
  fprintf(1,'maxiter exceeded.\n');
end

