function [A, b, fAb, min_eig, max_eig] = gen_data(which_matrix, f, opts)
%function [A, b, fAb, min_eig, max_eig] = gen_data(which_matrix, f, opts)
%
% GEN_DATA generates four types of matrices for testing. The actual
% returned matrix A is a replacement of the matrix one actually wants.
% It is a diagonal matrix which has the same spectrum as the wanted
% matrix. For example, if one want a lap2D matrix, the returned A has
% as diagonal entries the eigenvalues of a 2D Laplacian. The vector b
% is either the vector of all ones or a random vector. fAb is the
% ground truth f(A)*b. A has its extreme eigenvalues min_eig and
% max_eig.
%
% Matrices supported:
%
%  * 'rand'.    opts: 'm'
%  * 'uniform'. opts: 'm'
%  * 'lap2D'.   opts: 'm1', 'm2'
%  * 'covar'.   opts: 'm1', 'm2', 'alpha', 'j', 'deform', 'eigv'
%                     'alpha' should be less than min(m1,m2)-1
%                     choices for 'deform': 'none', 'scaley'
%

if ~exist('opts', 'var')
  opts = struct;
end

if strcmp(which_matrix, 'rand')

  m = setparam(opts, 'm', 1e4);

  dd = rand(m,1);
  A = spdiags(dd,0,m,m);
  
  b = randn(m,1);
  fAb = f(dd).*b;
  min_eig = min(dd);
  max_eig = max(dd);
  
elseif strcmp(which_matrix, 'uniform')

  m = setparam(opts, 'm', 1e4);

  dd = (1:m)'/m;
  A = spdiags(dd,0,m,m);
  
  b = ones(m,1);
  fAb = f(dd).*b;
  min_eig = min(dd);
  max_eig = max(dd);
  
elseif strcmp(which_matrix, 'lap2D')

  m1 = setparam(opts, 'm1', 100);
  m2 = setparam(opts, 'm2', 100);
  
  d1 = sin([1:m1]*pi/2/(m1+1)).^2;
  d2 = sin([1:m2]*pi/2/(m2+1)).^2;
  m = m1*m2;
  dd = 4 * reshape( repmat(d1,m2,1)+repmat(d2',1,m1), m, 1 );
  A = spdiags(dd,0,m,m);
  
  b = ones(m,1);
  fAb = f(dd).*b;
  min_eig = min(dd);
  max_eig = max(dd);

elseif strcmp(which_matrix, 'covar')

  m1 = setparam(opts, 'm1', 100);
  m2 = setparam(opts, 'm2', 100);
  alpha = setparam(opts, 'alpha', 12.5);
  j = setparam(opts, 'j', 3);
  deform = setparam(opts, 'deform', 'none');

  % window
  falpha = floor(alpha);
  window = repmat([-falpha:falpha]', 1, 2*falpha+1) + ...
           repmat([-falpha:falpha]*m1, 2*falpha+1, 1);

  % distances inside window
  x = repmat([-falpha:falpha], 2*falpha+1, 1);
  y = repmat([-falpha:falpha]', 1, 2*falpha+1);
  if strcmp(deform, 'scaley')
    suppx=max(max(x));
    inffx=min(min(x));
    mullx=2*(x-ones(size(x))*(suppx+inffx)/2)/(suppx-inffx);
    mullx=2*ones(size(x))-mullx.*mullx;
  else
    mullx = ones(size(y));
  end
  wdist = sqrt(x.^2+(mullx.*y).^2);
  
  % trim out the points farther than distance alpha
  window = window(:)';
  wdist = wdist(:)';
  x = x(:)';
  y = y(:)';
  ind = find(wdist<=alpha);
  window = window(ind);
  wdist = wdist(ind);
  x = x(ind);
  y = y(ind);

  % assemble idx, neighbors, and dist
  m = m1*m2;
  idx = repmat([1:m]', 1, length(window));
  neighbors = idx + repmat(window, m, 1);
  dist = repmat(wdist, m, 1);
  
  % trim out ghost neighbors
  %--north boarder
  for i = 1:falpha
    ind2 = find(y <= -i);
    ind1 = i + m1 * [0:(m2-1)];
    dist(ind1, ind2) = -1;
  end
  %--west boarder
  for i = 1:falpha
    ind2 = find(x <= -i);
    ind1 = [1:m1] + (i-1) * m1;
    dist(ind1, ind2) = -1;
  end
  %--south boarder
  for i = 1:falpha
    ind2 = find(y >= i);
    ind1 = [1:m2] * m1 - (i-1);
    dist(ind1, ind2) = -1;
  end
  %--east boarder
  for i = 1:falpha
    ind2 = find(x >= i);
    ind1 = m - ( [0:(m1-1)] + (i-1) * m1 );
    dist(ind1, ind2) = -1;
  end
  %--trim out ghost neighbors
  idx = idx(:);
  neighbors = neighbors(:);
  dist = dist(:);
  ind = find(dist>=0);
  idx = idx(ind);
  neighbors = neighbors(ind);
  dist = dist(ind);
  clear ind;
  
  % make the matrix A
  A = sparse(idx, neighbors, (1-dist./alpha).^(j+1), m, m);

  % the rest (only works for not large A)
  b = ones(m,1);
  dd = eig(full(A));
  A = spdiags(dd,0,m,m);
  fAb = f(dd).*b;
  min_eig = min(dd);
  max_eig = max(dd);
 
else
  
  error('unknown matrix');
  
end


%--------------------------------------------------------------
function x = setparam(opts, name, default)

if isfield(opts, name)
  x = opts.(name);
else
  x = default;
end
